/**
 * @file
 * @brief Object's transaction management interface.
 *
 * This structure defines the object's transaction management interface.
 */
/*
 * Copyright (C) 2015 Red Hat
 *
 * This file is part of tlog.
 *
 * Tlog is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Tlog is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with tlog; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef _TLOG_TRX_IFACE_H
#define _TLOG_TRX_IFACE_H

#include <tlog/trx_act.h>

/** An object's transaction management interface */
struct tlog_trx_iface {
    tlog_trx_act_fn     act;    /**< Transaction data transfer function */
};

#endif /* _TLOG_TRX_IFACE_H */
